from __future__ import annotations

from _thread import LockType, allocate_lock
from abc import ABC, abstractmethod
from typing import Protocol, TypeVar


class Waiter(Protocol):
    @abstractmethod
    def release(self) -> None:
        raise NotImplementedError


W = TypeVar("W", bound=Waiter)


def create_threading_waiter() -> LockType:
    # Create an already-acquired lock. This will be released by `release()`.
    lock = allocate_lock()
    lock.acquire()
    return lock


class AsyncWaiter(ABC):
    """
    Protocol used by :meth:`Logot.await_for` to pause tests until expected logs arrive.

    .. note::

        This class is for integration with :ref:`3rd-party asynchronous frameworks <integrations-async>`. It is not
        generally used when writing tests.
    """

    __slots__ = ()

    @abstractmethod
    def release(self) -> None:
        """
        Releases the test waiting on :meth:`wait`, allowing it to resume immediately.
        """
        raise NotImplementedError

    @abstractmethod
    async def wait(self, *, timeout: float) -> None:
        """
        Waits *asynchronously* for :meth:`release` to be called or the ``timeout`` to expire.

        :param timeout: How long to wait (in seconds) before resuming.
        """
        raise NotImplementedError



from __future__ import annotations

import trio
from trio.lowlevel import current_trio_token

from logot._wait import AsyncWaiter


class TrioWaiter(AsyncWaiter):
    """
    A :class:`logot.AsyncWaiter` implementation for :mod:`trio`.
    """

    __slots__ = ("_token", "_event")

    def __init__(self) -> None:
        self._token = current_trio_token()
        self._event = trio.Event()

    def release(self) -> None:
        self._token.run_sync_soon(self._event.set)

    async def wait(self, *, timeout: float) -> None:
        with trio.move_on_after(timeout):
            await self._event.wait()



from __future__ import annotations

from functools import partial

import loguru
from loguru import logger

from logot._capture import Captured
from logot._logot import Capturer, Logot
from logot._typing import Level, Name


class LoguruCapturer(Capturer):
    """
    A :class:`logot.Capturer` implementation for :mod:`loguru`.
    """

    __slots__ = ("_handler_id",)

    def start_capturing(self, logot: Logot, /, *, level: Level, name: Name) -> None:
        self._handler_id = logger.add(partial(_sink, logot=logot), level=level, filter=name)

    def stop_capturing(self) -> None:
        logger.remove(self._handler_id)


def _sink(msg: loguru.Message, *, logot: Logot) -> None:
    record = msg.record
    level = record["level"]
    logot.capture(Captured(level.name, record["message"], levelno=level.no))



from __future__ import annotations

import asyncio

from logot._wait import AsyncWaiter


class AsyncioWaiter(AsyncWaiter):
    """
    A :class:`logot.AsyncWaiter` implementation for :mod:`asyncio`.

    .. note::

        This is the default :class:`logot.AsyncWaiter` implementation.
    """

    __slots__ = ("_loop", "_future")

    def __init__(self) -> None:
        # Create an unresolved `asyncio.Future`. This will be resolved by `release()`.
        self._loop = asyncio.get_running_loop()
        self._future: asyncio.Future[None] = self._loop.create_future()

    def release(self) -> None:
        self._loop.call_soon_threadsafe(self._resolve)

    async def wait(self, *, timeout: float) -> None:
        timer = self._loop.call_later(timeout, self._resolve)
        try:
            await self._future
        finally:
            timer.cancel()

    def _resolve(self) -> None:
        try:
            self._future.set_result(None)
        except asyncio.InvalidStateError:  # pragma: no cover
            # It's possible that the timeout and the `release()` will both occur in the same tick of the event loop.
            pass



from __future__ import annotations

from logot._typing import Level, Name


def validate_level(level: Level) -> Level:
    # Handle `str` or `int` level.
    if isinstance(level, (str, int)):
        return level
    # Handle invalid level.
    raise TypeError(f"Invalid level: {level!r}")


def validate_name(name: Name) -> Name:
    # Handle `None` or `str` name.
    if name is None or isinstance(name, str):
        return name
    # Handle invalid name.
    raise TypeError(f"Invalid name: {name!r}")


def validate_timeout(timeout: float) -> float:
    # Handle numeric timeout.
    if isinstance(timeout, (float, int)):
        if timeout >= 0.0:
            return float(timeout)
        raise ValueError(f"Invalid timeout: {timeout!r}")
    # Handle invalid timeout.
    raise TypeError(f"Invalid timeout: {timeout!r}")



from __future__ import annotations

from abc import ABC, abstractmethod

from logot._capture import Captured
from logot._format import format_level, format_log
from logot._msg import compile_msg_matcher
from logot._typing import Level
from logot._validate import validate_level


class Logged(ABC):
    """
    A :doc:`log pattern </log-pattern-matching>` passed to :meth:`Logot.wait_for`, :meth:`Logot.await_for` and related
    APIs.

    .. important::

        This is an abstract class and cannot be instantiated. Use the helpers in :mod:`logot.logged` to create log
        patterns.

        :class:`Logged` instances are immutable and can be reused between tests.

    .. seealso::

        See :doc:`/log-pattern-matching` usage guide.
    """

    __slots__ = ()

    def __rshift__(self, logged: Logged) -> Logged:
        return _OrderedAllLogged.from_compose(self, logged)

    def __and__(self, logged: Logged) -> Logged:
        return _UnorderedAllLogged.from_compose(self, logged)

    def __or__(self, logged: Logged) -> Logged:
        return _AnyLogged.from_compose(self, logged)

    def __str__(self) -> str:
        return self._str(indent="")

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def _reduce(self, captured: Captured) -> Logged | None:
        raise NotImplementedError

    @abstractmethod
    def _str(self, *, indent: str) -> str:
        raise NotImplementedError


def log(level: Level, msg: str) -> Logged:
    """
    Creates a :doc:`log pattern </log-pattern-matching>` representing a log record at the given ``level`` with the given
    ``msg``.

    :param level: A log level name (e.g. ``"DEBUG"``) or numeric level (e.g. :data:`logging.DEBUG`).
    :param msg: A log :doc:`message pattern </log-message-matching>`.
    """
    return _RecordLogged(validate_level(level), msg)


def debug(msg: str) -> Logged:
    """
    Creates a :doc:`log pattern </log-pattern-matching>` representing a log record at ``DEBUG`` level with the given
    ``msg``.

    :param msg: A log :doc:`message pattern </log-message-matching>`.
    """
    return _RecordLogged("DEBUG", msg)


def info(msg: str) -> Logged:
    """
    Creates a :doc:`log pattern </log-pattern-matching>` representing a log record at ``INFO`` level with the given
    ``msg``.

    :param msg: A log :doc:`message pattern </log-message-matching>`.
    """
    return _RecordLogged("INFO", msg)


def warning(msg: str) -> Logged:
    """
    Creates a :doc:`log pattern </log-pattern-matching>` representing a log record at ``WARNING`` level with the given
    ``msg``.

    :param msg: A log :doc:`message pattern </log-message-matching>`.
    """
    return _RecordLogged("WARNING", msg)


def error(msg: str) -> Logged:
    """
    Creates a :doc:`log pattern </log-pattern-matching>` representing a log record at ``ERROR`` level with the given
    ``msg``.

    :param msg: A log :doc:`message pattern </log-message-matching>`.
    """
    return _RecordLogged("ERROR", msg)


def critical(msg: str) -> Logged:
    """
    Creates a :doc:`log pattern </log-pattern-matching>` representing a log record at ``CRITICAL`` level with the given
    ``msg``.

    :param msg: A log :doc:`message pattern </log-message-matching>`.
    """
    return _RecordLogged("CRITICAL", msg)


class _RecordLogged(Logged):
    __slots__ = ("_level", "_msg", "_msg_matcher")

    def __init__(self, level: Level, msg: str) -> None:
        self._level = level
        self._msg = msg
        self._msg_matcher = compile_msg_matcher(msg)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, _RecordLogged) and other._level == self._level and other._msg == self._msg

    def __repr__(self) -> str:
        return f"log({self._level!r}, {self._msg!r})"

    def _reduce(self, captured: Captured) -> Logged | None:
        # Match `str` level.
        if isinstance(self._level, str):
            if self._level != captured.levelname:
                return self
        # Match `int` level.
        elif isinstance(self._level, int):
            if self._level != captured.levelno:
                return self
        else:  # pragma: no cover
            raise TypeError(f"Invalid level: {self._level!r}")
        # Match message.
        if not self._msg_matcher(captured.msg):
            return self
        # We matched!
        return None

    def _str(self, *, indent: str) -> str:
        return format_log(format_level(self._level), self._msg)


class _ComposedLogged(Logged):
    __slots__ = ("_logged_items",)

    def __init__(self, logged_items: tuple[Logged, ...]) -> None:
        assert len(logged_items) > 1
        self._logged_items = logged_items

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and other._logged_items == self._logged_items

    @classmethod
    def from_compose(cls, logged_a: Logged, logged_b: Logged) -> Logged:
        # If possible, flatten nested logged items of the same type.
        if isinstance(logged_a, cls):
            if isinstance(logged_b, cls):
                return cls((*logged_a._logged_items, *logged_b._logged_items))
            return cls((*logged_a._logged_items, logged_b))
        if isinstance(logged_b, cls):
            return cls((logged_a, *logged_b._logged_items))
        # Wrap the logged items without flattening.
        return cls((logged_a, logged_b))

    @classmethod
    def from_reduce(cls, logged_items: tuple[Logged, ...]) -> Logged | None:
        assert logged_items
        # If there is a single logged item, do not wrap it.
        if len(logged_items) == 1:
            return logged_items[0]
        # Wrap the logged items.
        return cls(logged_items)


class _OrderedAllLogged(_ComposedLogged):
    __slots__ = ()

    def __repr__(self) -> str:
        return f"({' >> '.join(map(repr, self._logged_items))})"

    def _reduce(self, captured: Captured) -> Logged | None:
        logged = self._logged_items[0]
        reduced = logged._reduce(captured)
        # Handle full reduction.
        if reduced is None:
            return _OrderedAllLogged.from_reduce(self._logged_items[1:])
        # Handle partial reduction.
        if reduced is not logged:
            return _OrderedAllLogged((reduced, *self._logged_items[1:]))
        # Handle no reduction.
        return self

    def _str(self, *, indent: str) -> str:
        return f"\n{indent}".join(logged._str(indent=indent) for logged in self._logged_items)


class _UnorderedAllLogged(_ComposedLogged):
    __slots__ = ()

    def __repr__(self) -> str:
        return f"({' & '.join(map(repr, self._logged_items))})"

    def _reduce(self, captured: Captured) -> Logged | None:
        for n, logged in enumerate(self._logged_items):
            reduced = logged._reduce(captured)
            # Handle full reduction.
            if reduced is None:
                return _UnorderedAllLogged.from_reduce((*self._logged_items[:n], *self._logged_items[n + 1 :]))
            # Handle partial reduction.
            if reduced is not logged:
                return _UnorderedAllLogged((*self._logged_items[:n], reduced, *self._logged_items[n + 1 :]))
        # Handle no reduction.
        return self

    def _str(self, *, indent: str) -> str:
        nested_indent = indent + "  "
        logged_items_str = "".join(f"\n{indent}- {logged._str(indent=nested_indent)}" for logged in self._logged_items)
        return f"Unordered:{logged_items_str}"


class _AnyLogged(_ComposedLogged):
    __slots__ = ()

    def __repr__(self) -> str:
        return f"({' | '.join(map(repr, self._logged_items))})"

    def _reduce(self, captured: Captured) -> Logged | None:
        for n, logged in enumerate(self._logged_items):
            reduced = logged._reduce(captured)
            # Handle full reduction.
            if reduced is None:
                return None
            # Handle partial reduction.
            if reduced is not logged:
                return _AnyLogged((*self._logged_items[:n], reduced, *self._logged_items[n + 1 :]))
        # Handle no reduction.
        return self

    def _str(self, *, indent: str) -> str:
        nested_indent = indent + "  "
        logged_items_str = "".join(f"\n{indent}- {logged._str(indent=nested_indent)}" for logged in self._logged_items)
        return f"Any:{logged_items_str}"



"""
Log :doc:`pattern matching </log-pattern-matching>` API.

.. seealso::

    See :doc:`/log-pattern-matching` usage guide.
"""
from __future__ import annotations

from logot._logged import critical as critical
from logot._logged import debug as debug
from logot._logged import error as error
from logot._logged import info as info
from logot._logged import log as log
from logot._logged import warning as warning



"""
Integration API for :mod:`loguru`.

.. seealso::

    See :doc:`/integrations/loguru` usage guide.
"""
from __future__ import annotations

from logot._loguru import LoguruCapturer as LoguruCapturer



from __future__ import annotations

from typing import Callable, ClassVar
from unittest import TestCase, TestResult

from logot._logot import Capturer, Logot
from logot._typing import Level, Name
from logot._wait import AsyncWaiter


class LogotTestCase(TestCase):
    """
    A :class:`unittest.TestCase` subclass with automatic :doc:`log capturing </log-capturing>`.
    """

    logot: Logot
    """
    An initialized :class:`logot.Logot` instance with :doc:`log capturing </log-capturing>` enabled.

    Use this to make log assertions in your tests.
    """

    logot_level: ClassVar[Level] = Logot.DEFAULT_LEVEL
    """
    The ``level`` used for automatic :doc:`log capturing </log-capturing>`.

    Defaults to :attr:`logot.Logot.DEFAULT_LEVEL`.
    """

    logot_name: ClassVar[Name] = Logot.DEFAULT_NAME
    """
    The ``name`` used for automatic :doc:`log capturing </log-capturing>`.

    Defaults to :attr:`logot.Logot.DEFAULT_NAME`.
    """

    logot_capturer: ClassVar[Callable[[], Capturer]] = Logot.DEFAULT_CAPTURER
    """
    The default ``capturer`` for :attr:`LogotTestCase.logot`.

    Defaults to :attr:`logot.Logot.DEFAULT_CAPTURER`.
    """

    logot_timeout: ClassVar[float] = Logot.DEFAULT_TIMEOUT
    """
    The default ``timeout`` (in seconds) for :attr:`LogotTestCase.logot`.

    Defaults to :attr:`logot.Logot.DEFAULT_TIMEOUT`.
    """

    logot_async_waiter: ClassVar[Callable[[], AsyncWaiter]] = Logot.DEFAULT_ASYNC_WAITER
    """
    The default ``async_waiter`` for :attr:`LogotTestCase.logot`.

    Defaults to :attr:`logot.Logot.DEFAULT_ASYNC_WAITER`.
    """

    def _logot_setup(self) -> None:
        self.logot = Logot(
            capturer=self.__class__.logot_capturer,
            timeout=self.__class__.logot_timeout,
            async_waiter=self.__class__.logot_async_waiter,
        )
        # TODO: Use `TestCase.enterContext()` when we only need to support Python 3.11+.
        ctx = self.logot.capturing(level=self.logot_level, name=self.logot_name)
        ctx.__enter__()
        self.addCleanup(ctx.__exit__, None, None, None)

    def run(self, result: TestResult | None = None) -> TestResult | None:
        self._logot_setup()
        return super().run(result)

    def debug(self) -> None:
        self._logot_setup()
        return super().debug()



from __future__ import annotations

from collections.abc import Generator
from typing import Callable

import pytest

from logot._import import import_any_parsed
from logot._logot import Capturer, Logot
from logot._typing import MISSING, Level, Name, T
from logot._wait import AsyncWaiter


def pytest_addoption(parser: pytest.Parser, pluginmanager: pytest.PytestPluginManager) -> None:
    group = parser.getgroup("logot")
    _add_option(
        parser,
        group,
        name="level",
        help="The `level` used for automatic `logot` log capturing",
    )
    _add_option(
        parser,
        group,
        name="name",
        help="The `name` used for automatic `logot` log capturing",
    )
    _add_option(
        parser,
        group,
        name="capturer",
        help="The default `capturer` for the `logot` fixture",
    )
    _add_option(
        parser,
        group,
        name="timeout",
        help="The default `timeout` (in seconds) for the `logot` fixture",
    )
    _add_option(
        parser,
        group,
        name="async_waiter",
        help="The default `async_waiter` for the `logot` fixture",
    )


@pytest.fixture()
def logot(
    logot_level: Level,
    logot_name: Name,
    logot_capturer: Callable[[], Capturer],
    logot_timeout: float,
    logot_async_waiter: Callable[[], AsyncWaiter],
) -> Generator[Logot, None, None]:
    """
    An initialized `logot.Logot` instance with log capturing enabled.
    """
    logot = Logot(capturer=logot_capturer, timeout=logot_timeout, async_waiter=logot_async_waiter)
    with logot.capturing(level=logot_level, name=logot_name):
        yield logot


@pytest.fixture(scope="session")
def logot_level(request: pytest.FixtureRequest) -> Level:
    """
    The `level` used for automatic log capturing.
    """
    return _get_option(request, name="level", parser=str, default=Logot.DEFAULT_LEVEL)


@pytest.fixture(scope="session")
def logot_name(request: pytest.FixtureRequest) -> Name:
    """
    The `name` used for automatic log capturing.
    """
    return _get_option(request, name="name", parser=str, default=Logot.DEFAULT_NAME)


@pytest.fixture(scope="session")
def logot_capturer(request: pytest.FixtureRequest) -> Callable[[], Capturer]:
    """
    The default `capturer` for the `logot` fixture.
    """
    return _get_option(request, name="capturer", parser=import_any_parsed, default=Logot.DEFAULT_CAPTURER)


@pytest.fixture(scope="session")
def logot_timeout(request: pytest.FixtureRequest) -> float:
    """
    The default `timeout` (in seconds) for the `logot` fixture.
    """
    return _get_option(request, name="timeout", parser=float, default=Logot.DEFAULT_TIMEOUT)


@pytest.fixture(scope="session")
def logot_async_waiter(request: pytest.FixtureRequest) -> Callable[[], AsyncWaiter]:
    """
    The default `async_waiter` for the `logot` fixture.
    """
    return _get_option(request, name="async_waiter", parser=import_any_parsed, default=Logot.DEFAULT_ASYNC_WAITER)


def get_qualname(name: str) -> str:
    return f"logot_{name}"


def get_optname(name: str) -> str:
    return f"--logot-{name.replace('_', '-')}"


def _add_option(parser: pytest.Parser, group: pytest.OptionGroup, *, name: str, help: str) -> None:
    qualname = get_qualname(name)
    parser.addini(qualname, default=MISSING, help=help)
    group.addoption(
        get_optname(name),
        default=MISSING,
        dest=qualname,
        metavar=name.upper(),
        help=help,
    )


def _get_option(request: pytest.FixtureRequest, *, name: str, parser: Callable[[str], T], default: T) -> T:
    qualname = get_qualname(name)
    # Try to get the value from the command line, followed by the config file.
    value: str = request.config.getoption(qualname, default=MISSING)
    if value is MISSING:
        value = request.config.getini(qualname)
        if value is MISSING:
            # Give up and return the default.
            return default
    # Parse and return the value.
    try:
        return parser(value)
    except Exception as ex:
        raise pytest.UsageError(f"Invalid {qualname}: {ex}") from ex



from __future__ import annotations

import sys
from typing import Any, TypeVar, Union

if sys.version_info >= (3, 10):
    from typing import ParamSpec as ParamSpec
    from typing import TypeAlias as TypeAlias
else:
    from typing_extensions import ParamSpec as ParamSpec
    from typing_extensions import TypeAlias as TypeAlias

P = ParamSpec("P")
T = TypeVar("T")

# TODO: Use `UnionType` when we only need to support Python 3.10+.
Level: TypeAlias = Union[str, int]
Name: TypeAlias = Union[str, None]

MISSING: Any = object()



from __future__ import annotations

from importlib import import_module
from typing import Any, Callable


def import_any(module: str, name: str) -> Any:
    module_obj = import_module(module)
    return getattr(module_obj, name)


def import_any_parsed(name: str) -> Any:
    module, name = name.rsplit(".", 1)
    return import_any(module, name)


class LazyCallable:
    __slots__ = ("_module", "_name", "_resolved")

    _resolved: Callable[..., Any]

    def __init__(self, module: str, name: str) -> None:
        self._module = module
        self._name = name

    def __call__(self, /, *args: Any, **kwargs: Any) -> Any:
        try:
            resolved = self._resolved
        except AttributeError:
            resolved = self._resolved = import_any(self._module, self._name)
        return resolved(*args, **kwargs)

    def __repr__(self) -> str:
        return f"{self._module}.{self._name}"



"""
Integration API for :mod:`logging`.

.. seealso::

    See :doc:`/log-capturing` usage guide.
"""
from __future__ import annotations

from logot._logging import LoggingCapturer as LoggingCapturer



from __future__ import annotations

import re
from typing import Callable

from logot._typing import TypeAlias

# Compiled matcher callable.
# The returned `object` is truthy on successful match and falsy on failed match.
MessageMatcher: TypeAlias = Callable[[str], object]

# Regex matching a simplified conversion specifier.
_RE_CONVERSION = re.compile(r"%(.|$)")

# Mapping of conversion types to regex matchers.
_CONVERSION_INT = r"\-?\d+"
_CONVERSION_STR = r".*?"
_CONVERSION_MAP = {
    # Integer conversion.
    "d": _CONVERSION_INT,
    "i": _CONVERSION_INT,
    "o": r"\-?[0-8]+",
    "u": _CONVERSION_INT,
    "x": r"\-?[\da-f]+",
    "X": r"\-?[\dA-F]+",
    # Float conversion.
    "e": r"\-?(?:\d+\.\d+e[\-\+]\d{2,3}|inf|nan)",
    "E": r"\-?(?:\d+\.\d+E[\-\+]\d{2,3}|INF|NAN)",
    "f": r"\-?(?:\d+\.\d+|inf|nan)",
    "F": r"\-?(?:\d+\.\d+|INF|NAN)",
    "g": r"\-?(?:\d+(?:\.\d+)?(?:e[\-\+]\d{2,3})?|inf|nan)",
    "G": r"\-?(?:\d+(?:\.\d+)?(?:E[\-\+]\d{2,3})?|INF|NAN)",
    # Character conversion.
    "c": r".",
    # String conversion.
    "r": _CONVERSION_STR,
    "s": _CONVERSION_STR,
    "a": _CONVERSION_STR,
    # Percent conversion.
    "%": r"%",
}


def compile_msg_matcher(pattern: str) -> MessageMatcher:
    parts: list[str] = _RE_CONVERSION.split(pattern)
    parts_len = len(parts)
    # If there is more than one part, at least one conversion specifier was found and we might need a regex matcher.
    if parts_len > 1:
        is_regex = False
        # Replace conversion types with regex matchers.
        for n in range(1, parts_len, 2):
            part = parts[n]
            try:
                parts[n] = _CONVERSION_MAP[part]
            except KeyError:
                part_index = sum(map(len, parts[:n:2])) + ((n // 2) * 2) + 1
                raise ValueError(f"Unsupported format character {part!r} at index {part_index}") from None
            # A "%" is used as an escape sequence, and doesn't require a regex matcher. Anything else does.
            is_regex |= part != "%"
        # Create regex matcher.
        if is_regex:
            parts[::2] = map(re.escape, parts[::2])
            return re.compile("".join(parts), re.DOTALL).fullmatch
        # Recreate the pattern with all escape sequences replaced.
        pattern = "".join(parts)
    # Create simple matcher.
    return pattern.__eq__



from __future__ import annotations

from logot._typing import Level


def format_level(level: Level) -> str:
    # Format `str` level.
    if isinstance(level, str):
        return level
    # Format `int` level.
    if isinstance(level, int):
        return f"Level {level}"
    # Handle invalid level.
    raise TypeError(f"Invalid level: {level!r}")


def format_log(levelname: str, msg: str) -> str:
    return f"[{levelname}] {msg}"



from __future__ import annotations

from logot._format import format_log


class Captured:
    """
    A captured log record.

    Send :class:`Captured` logs to :meth:`Logot.capture` to integrate with
    :ref:`3rd-party logging frameworks <integrations-logging>`.

    .. note::

        This class is for integration with :ref:`3rd-party logging frameworks <integrations-logging>`. It is not
        generally used when writing tests.

    :param levelname: See :attr:`Captured.levelname`.
    :param msg: See :attr:`Captured.msg`.
    :param levelno: See :attr:`Captured.levelno`.
    """

    __slots__ = ("levelname", "msg", "levelno")

    levelname: str
    """
    The log level name (e.g. ``"DEBUG"``).
    """

    msg: str
    """
    The log message.
    """

    levelno: int | None
    """
    The log level number (e.g. :data:`logging.DEBUG`).

    This is an *optional* log capture field. When provided, it allows matching
    :doc:`log patterns </log-pattern-matching>` from :func:`logged.log` with a numeric ``level``.
    """

    def __init__(self, levelname: str, msg: str, *, levelno: int | None = None) -> None:
        self.levelname = levelname
        self.msg = msg
        self.levelno = levelno

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Captured)
            and other.levelname == self.levelname
            and other.msg == self.msg
            and other.levelno == self.levelno
        )

    def __repr__(self) -> str:
        return f"Captured({self.levelname!r}, {self.msg!r}, levelno={self.levelno!r})"

    def __str__(self) -> str:
        return format_log(self.levelname, self.msg)



"""
Main :mod:`logot` API for log-based testing.

.. seealso::

    See :doc:`/index` usage guide.
"""
from __future__ import annotations

from logot._capture import Captured as Captured
from logot._logged import Logged as Logged
from logot._logot import Capturer as Capturer
from logot._logot import Logot as Logot
from logot._wait import AsyncWaiter as AsyncWaiter



"""
Integration API for :mod:`asyncio`.

.. seealso::

    See :ref:`index-testing-async` usage guide.
"""
from __future__ import annotations

from logot._asyncio import AsyncioWaiter as AsyncioWaiter



from __future__ import annotations

import logging

from logot._capture import Captured
from logot._logot import Capturer, Logot
from logot._typing import Level, Name


class LoggingCapturer(Capturer):
    """
    A :class:`logot.Capturer` implementation for :mod:`logging`.

    .. note::

        This is the default :class:`logot.Capturer` implementation.
    """

    __slots__ = ("_logger", "_handler", "_prev_levelno")

    def start_capturing(self, logot: Logot, /, *, level: Level, name: Name) -> None:
        logger = self._logger = logging.getLogger(name)
        handler = self._handler = _Handler(level, logot)
        # If the logger is less verbose than the handler, force it to the necessary verboseness.
        self._prev_levelno = logger.level
        if handler.level < logger.getEffectiveLevel():
            logger.setLevel(handler.level)
        # Add the handler.
        logger.addHandler(handler)

    def stop_capturing(self) -> None:
        # Remove the handler and restore the previous level.
        self._logger.removeHandler(self._handler)
        self._logger.setLevel(self._prev_levelno)


class _Handler(logging.Handler):
    __slots__ = ("_logot",)

    def __init__(self, level: Level, logot: Logot) -> None:
        super().__init__(level)
        self._logot = logot

    def emit(self, record: logging.LogRecord) -> None:
        captured = Captured(record.levelname, record.getMessage(), levelno=record.levelno)
        self._logot.capture(captured)



from __future__ import annotations

from _thread import allocate_lock
from abc import ABC, abstractmethod
from collections import deque
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Any, Callable, ClassVar, Generic

from logot._capture import Captured
from logot._import import LazyCallable
from logot._logged import Logged
from logot._typing import Level, Name
from logot._validate import validate_level, validate_name, validate_timeout
from logot._wait import AsyncWaiter, W, create_threading_waiter


class Logot:
    """
    The main :mod:`logot` API for capturing and waiting for logs.

    .. seealso::

        See :doc:`/index` usage guide.

    :param capturer: See :attr:`Logot.capturer`.
    :param timeout: See :attr:`Logot.timeout`.
    :param async_waiter: See :attr:`Logot.async_waiter`.
    """

    __slots__ = ("capturer", "timeout", "async_waiter", "_lock", "_queue", "_wait")

    DEFAULT_LEVEL: ClassVar[Level] = "DEBUG"
    """
    The default ``level`` used by :meth:`capturing`.
    """

    DEFAULT_NAME: ClassVar[Name] = None
    """
    The default ``name`` used by :meth:`capturing`.

    Defaults to :data:`None`, representing the root logger.
    """

    DEFAULT_CAPTURER: ClassVar[Callable[[], Capturer]] = LazyCallable("logot.logging", "LoggingCapturer")
    """
    The default :attr:`capturer` for new :class:`Logot` instances.
    """

    DEFAULT_TIMEOUT: ClassVar[float] = 3.0
    """
    The default :attr:`timeout` (in seconds) for new :class:`Logot` instances.
    """

    DEFAULT_ASYNC_WAITER: ClassVar[Callable[[], AsyncWaiter]] = LazyCallable("logot.asyncio", "AsyncioWaiter")
    """
    The default :attr:`async_waiter` for new :class:`Logot` instances.
    """

    capturer: Callable[[], Capturer]
    """
    The default ``capturer`` used by :meth:`capturing`.

    .. note::

        This is for integration with :ref:`3rd-party logging frameworks <integrations-logging>`.

    Defaults to :attr:`Logot.DEFAULT_CAPTURER`.
    """

    timeout: float
    """
    The default ``timeout`` (in seconds) for calls to :meth:`wait_for` and :meth:`await_for`.

    Defaults to :attr:`Logot.DEFAULT_TIMEOUT`.
    """

    async_waiter: Callable[[], AsyncWaiter]
    """
    The default ``async_waiter`` for calls to :meth:`await_for`.

    .. note::

        This is for integration with :ref:`3rd-party asynchronous frameworks <integrations-async>`.

    Defaults to :attr:`Logot.DEFAULT_ASYNC_WAITER`.
    """

    def __init__(
        self,
        *,
        capturer: Callable[[], Capturer] = DEFAULT_CAPTURER,
        timeout: float = DEFAULT_TIMEOUT,
        async_waiter: Callable[[], AsyncWaiter] = DEFAULT_ASYNC_WAITER,
    ) -> None:
        self.capturer = capturer
        self.timeout = validate_timeout(timeout)
        self.async_waiter = async_waiter
        self._lock = allocate_lock()
        self._queue: deque[Captured] = deque()
        self._wait: _Wait[Any] | None = None

    def capturing(
        self,
        *,
        level: Level = DEFAULT_LEVEL,
        name: Name = DEFAULT_NAME,
        capturer: Callable[[], Capturer] | None = None,
    ) -> AbstractContextManager[Logot]:
        """
        Captures logs emitted at the given ``level`` (or higher) by the given logger ``name`` for the duration of the
        context.

        If the named logger level is less verbose than the requested ``level``, it will be temporarily adjusted to the
        requested ``level`` for the duration of the context.

        .. seealso::

            See :doc:`/log-capturing` usage guide.

        :param level: A log level name (e.g. ``"DEBUG"``) or numeric level (e.g. :data:`logging.DEBUG`). Defaults to
            :attr:`Logot.DEFAULT_LEVEL`.
        :param name: A logger name to capture logs from. Defaults to :attr:`Logot.DEFAULT_NAME`.
        :param capturer: Protocol used to capture logs. This is for integration with
            :ref:`3rd-party logging frameworks <integrations-logging>`. Defaults to :attr:`Logot.capturer`.
        """
        if capturer is None:
            capturer = self.capturer
        capturer_obj = capturer()
        level = validate_level(level)
        name = validate_name(name)
        return _Capturing(self, capturer_obj, level=level, name=name)

    def capture(self, captured: Captured) -> None:
        """
        Adds the given captured log record to the internal capture queue.

        Any waiters blocked on :meth:`wait_for` to :meth:`await_for` will be notified and wake up if their
        :doc:`log pattern </log-pattern-matching>` is satisfied.

        .. note::

            This method is for integration with :ref:`3rd-party logging frameworks <integrations-logging>`. It is not
            generally used when writing tests.

        :param captured: The captured log.
        """
        with self._lock:
            # If there is a waiter that has not been fully reduced, attempt to reduce it.
            if self._wait is not None and self._wait.logged is not None:
                self._wait.logged = self._wait.logged._reduce(captured)
                # If the waiter has fully reduced, release the blocked caller.
                if self._wait.logged is None:
                    self._wait.waiter_obj.release()
                return
            # Otherwise, buffer the captured log.
            self._queue.append(captured)

    def assert_logged(self, logged: Logged) -> None:
        """
        Fails *immediately* if the expected ``log`` pattern has not arrived.

        :param logged: The expected :doc:`log pattern </log-pattern-matching>`.
        :raises AssertionError: If the expected ``log`` pattern has not arrived.
        """
        reduced = self._reduce(logged)
        if reduced is not None:
            raise AssertionError(f"Not logged:\n\n{reduced}")

    def assert_not_logged(self, logged: Logged) -> None:
        """
        Fails *immediately* if the expected ``log`` pattern **has** arrived.

        :param logged: The expected :doc:`log pattern </log-pattern-matching>`.
        :raises AssertionError: If the expected ``log`` pattern **has** arrived.
        """
        reduced = self._reduce(logged)
        if reduced is None:
            raise AssertionError(f"Logged:\n\n{logged}")

    def wait_for(
        self,
        logged: Logged,
        *,
        timeout: float | None = None,
    ) -> None:
        """
        Waits for the expected ``log`` pattern to arrive or the ``timeout`` to expire.

        :param logged: The expected :doc:`log pattern </log-pattern-matching>`.
        :param timeout: How long to wait (in seconds) before failing the test. Defaults to :attr:`Logot.timeout`.
        :raises AssertionError: If the expected ``log`` pattern does not arrive within ``timeout`` seconds.
        """
        wait = self._start_waiting(logged, create_threading_waiter, timeout=timeout)
        if wait is None:
            return
        try:
            wait.waiter_obj.acquire(timeout=wait.timeout)
        finally:
            self._stop_waiting(wait)

    async def await_for(
        self,
        logged: Logged,
        *,
        timeout: float | None = None,
        async_waiter: Callable[[], AsyncWaiter] | None = None,
    ) -> None:
        """
        Waits *asynchronously* for the expected ``log`` pattern to arrive or the ``timeout`` to expire.

        :param logged: The expected :doc:`log pattern </log-pattern-matching>`.
        :param timeout: How long to wait (in seconds) before failing the test. Defaults to :attr:`Logot.timeout`.
        :param async_waiter: Protocol used to pause tests until expected logs arrive. This is for integration with
            :ref:`3rd-party asynchronous frameworks <integrations-async>`. Defaults to :attr:`Logot.async_waiter`.
        :raises AssertionError: If the expected ``log`` pattern does not arrive within ``timeout`` seconds.
        """
        if async_waiter is None:
            async_waiter = self.async_waiter
        wait = self._start_waiting(logged, async_waiter, timeout=timeout)
        if wait is None:
            return
        try:
            await wait.waiter_obj.wait(timeout=wait.timeout)
        finally:
            self._stop_waiting(wait)

    def clear(self) -> None:
        """
        Clears any captured logs.
        """
        self._queue.clear()

    def _start_waiting(
        self, logged: Logged | None, waiter: Callable[[], W], *, timeout: float | None
    ) -> _Wait[W] | None:
        with self._lock:
            # If no timeout is provided, use the default timeout.
            # Otherwise, validate and use the provided timeout.
            if timeout is None:
                timeout = self.timeout
            else:
                timeout = validate_timeout(timeout)
            # Ensure no other waiters.
            if self._wait is not None:  # pragma: no cover
                raise RuntimeError("Multiple concurrent waiters are not supported")
            # Apply an immediate reduction.
            logged = self._reduce(logged)
            if logged is None:
                return None
            # All done!
            waiter_obj = waiter()
            wait = self._wait = _Wait(logged=logged, timeout=timeout, waiter_obj=waiter_obj)
            return wait

    def _stop_waiting(self, wait: _Wait[Any]) -> None:
        with self._lock:
            # Clear the waiter.
            self._wait = None
            # Error if the waiter logs are not fully reduced.
            if wait.logged is not None:
                raise AssertionError(f"Not logged:\n\n{wait.logged}")

    def _reduce(self, logged: Logged | None) -> Logged | None:
        # Drain the queue until the log is fully reduced.
        # This does not need a lock, since `deque.popleft()` is thread-safe.
        while logged is not None:
            try:
                captured = self._queue.popleft()
            except IndexError:
                break
            logged = logged._reduce(captured)
        # All done!
        return logged

    def __repr__(self) -> str:
        return f"Logot(capturer={self.capturer!r}, timeout={self.timeout!r}, async_waiter={self.async_waiter!r})"


class Capturer(ABC):
    """
    Protocol used by :meth:`Logot.capturing` to capture logs.

    .. note::

        This class is for integration with :ref:`3rd-party logging frameworks <integrations-logging>`. It is not
        generally used when writing tests.
    """

    __slots__ = ()

    @abstractmethod
    def start_capturing(self, logot: Logot, /, *, level: Level, name: Name) -> None:
        """
        Starts capturing logs for the given :class:`Logot`.

        Captured logs should be converted to a :class:`Captured` instance and sent to :meth:`Logot.capture`.

        :param logot: The :class:`Logot` instance.
        :param level: A log level name (e.g. ``"DEBUG"``) or numeric level (e.g. :data:`logging.DEBUG`).
        :param name: A logger name to capture logs from.
        """
        raise NotImplementedError

    @abstractmethod
    def stop_capturing(self) -> None:
        """
        Stops capturing logs.
        """
        raise NotImplementedError


class _Capturing:
    __slots__ = ("_logot", "_capturer_obj", "_level", "_name")

    def __init__(self, logot: Logot, capturer: Capturer, *, level: Level, name: Name) -> None:
        self._logot = logot
        self._capturer_obj = capturer
        self._level = level
        self._name = name

    def __enter__(self) -> Logot:
        self._capturer_obj.start_capturing(self._logot, level=self._level, name=self._name)
        return self._logot

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        self._capturer_obj.stop_capturing()


class _Wait(Generic[W]):
    __slots__ = ("logged", "timeout", "waiter_obj")

    def __init__(self, *, logged: Logged | None, timeout: float, waiter_obj: W) -> None:
        self.logged = logged
        self.timeout = timeout
        self.waiter_obj = waiter_obj
